//
// Sample to OpenGL cubic maps
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	<windows.h>
#include	<stdio.h>
#include	<string.h>
#include	<gl/gl.h>
#include	<gl/glut.h>
#include	<gl/glaux.h>
#include	"../glext.h"

#include	"Vector3D.h"
#include	"Vector2D.h"

extern "C"
{
	#include	"tga.h"
}

bool	loadTexture    ( int target, bool mipmap, const char * fileName );
bool	loadTgaTexture ( int target, bool mipmap, const char * fileName );
bool	loadBmpTexture ( int target, bool mipmap, const char * fileName );

float	 angle    = 0;
float	 angle2   = 0;
int		 reflMode = GL_REFLECTION_MAP_ARB;
int		 object   = 0;
unsigned texture;

unsigned	createTexture2D ( bool mipmap, const char * fileName )
{
	unsigned	textureId;

	glGenTextures   ( 1, &textureId );
	glBindTexture   ( GL_TEXTURE_2D, textureId );
	glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );							// set 1-byte alignment

	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );	// set texture to repeat mode
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

	loadTexture ( GL_TEXTURE_2D, mipmap, fileName );

	glTexParameteri   ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );	
	glTexParameteri   ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );

	return textureId;
}

unsigned	createCubeMap ( bool mipmap, const char * maps [] )
{
	unsigned	textureId;

	glGenTextures   ( 1, &textureId );
	glBindTexture   ( GL_TEXTURE_CUBE_MAP_ARB, textureId );
	glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );							// set 1-byte alignment

	glTexParameterf ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameterf ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	loadTexture     ( GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, mipmap, maps [0] );
	loadTexture     ( GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, mipmap, maps [1] );
	loadTexture     ( GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, mipmap, maps [2] );
	loadTexture     ( GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, mipmap, maps [3] );
	loadTexture     ( GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, mipmap, maps [4] );
	loadTexture     ( GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, mipmap, maps [5] );

	glEnable ( GL_TEXTURE_CUBE_MAP_ARB );

	if ( mipmap )
	{
		glTexParameteri ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
		glTexParameteri ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	} 
	else 
	{
		glTexParameteri ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	}

	return textureId;
}

bool	loadTexture ( int target, bool mipmap, const char * fileName )
{
	const char * ext = strrchr ( fileName, '.' );

	if ( ext == NULL )
		return false;

	if ( !stricmp ( ext + 1, "bmp" ) )
		return loadBmpTexture ( target, mipmap, fileName );
	else
	if ( !stricmp ( ext + 1, "tga" ) )
		return loadTgaTexture ( target, mipmap, fileName );

	printf ( "oops\n" );

	return false;
}

bool	loadTgaTexture ( int target, bool mipmap, const char * fileName )
{
	FILE            * file;
	gliGenericImage * image;

	file = fopen ( fileName, "rb" );

	if ( file == NULL )
	{
		printf ( "Error loading %s.\n", fileName );

		return false;
	}

	image = gliReadTGA ( file, (char *)fileName );

	fclose ( file );

	if ( mipmap )
		gluBuild2DMipmaps ( target, image -> components,
			                image -> width, image -> height,
							image -> format, GL_UNSIGNED_BYTE, image -> pixels );
	else
	    glTexImage2D (      target, 0, image -> components,
				            image -> width, image -> height, 0,
							image -> format, GL_UNSIGNED_BYTE, image -> pixels );
	return true;
}

bool	loadBmpTexture ( int target, bool mipmap, const char * fileName )
{
								// load texture
	AUX_RGBImageRec * texture = auxDIBImageLoad ( fileName );

	if ( texture == NULL )
	{
		printf ( "Error loading %s.\n", fileName );

		return false;
	}

	if ( mipmap )
		gluBuild2DMipmaps ( target, GL_RGB, 
			                texture -> sizeX, texture -> sizeY, 
				            GL_RGB, GL_UNSIGNED_BYTE, 
					        texture -> data );
	else
	    glTexImage2D ( target, 0, GL_RGB,
					   texture -> sizeX, texture -> sizeY, 0,
					   GL_RGB, GL_UNSIGNED_BYTE, texture -> data );


	return true;
}

bool	isExtensionSupported ( const char * ext )
{
	const char * extensions = (const char *)glGetString ( GL_EXTENSIONS );
	const char * start      = extensions;
	const char * ptr;

	while ( ( ptr = strstr ( start, ext ) ) != NULL )
	{
											// we've found, ensure name is exactly ext
		const char * end = ptr + strlen ( ext );

		if ( isspace ( *end ) || *end == '\0' )
			return true;

		start = end;
	}

	return false;
}

void init ()
{
	glClearColor ( 0, 0, 0, 1 );
	glEnable     ( GL_DEPTH_TEST );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glTexGeni ( GL_S, GL_TEXTURE_GEN_MODE, reflMode );
    glTexGeni ( GL_T, GL_TEXTURE_GEN_MODE, reflMode );
	glTexGeni ( GL_R, GL_TEXTURE_GEN_MODE, reflMode );

	glEnable  ( GL_TEXTURE_CUBE_MAP_ARB );

	glBindTexture ( GL_TEXTURE_CUBE_MAP_ARB, texture );

	glTexParameteri ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri ( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP );

    glEnable  ( GL_TEXTURE_GEN_S );
    glEnable  ( GL_TEXTURE_GEN_T );
    glEnable  ( GL_TEXTURE_GEN_R );

	glPushMatrix    ();
	glTranslatef    ( 2, 2, 2 );
	glRotatef       ( angle,  1, 1, 0 );
	glRotatef       ( angle2, 0, 1, 1 );

	if ( object == 0 )
		glutSolidTeapot ( 1 );
	else
	if ( object == 1 )
		glutSolidSphere ( 1, 30, 30 );
	else
	if ( object == 2 )
		glutSolidCube ( 1 );
	else
	if ( object == 3 )
		glutSolidCone ( 0.7, 1.2, 30, 30 );
	else
	if ( object == 4 )
		glutSolidTorus ( 0.5, 1, 40, 40 );
	else
	if ( object == 5 )
		glutSolidDodecahedron ();
	else
	if ( object == 6 )
		glutSolidOctahedron ();
	else
	if ( object == 7 )
		glutSolidTetrahedron ();
	else
	if ( object == 8 )
		glutSolidIcosahedron ();

	glPopMatrix     ();

	glutSwapBuffers ();
}

void reshape ( int w, int h ) 
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( 0.0, 0.0, 0.0,		// eye
					5.0, 5.0, 5.0,		// center
					0.0, 1.0, 0.0 );
}


void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )		//quit requested
		exit ( 0 );

	if ( key == 'n' || key == 'N' )
		reflMode = GL_NORMAL_MAP_ARB;
	else
	if ( key == 'r' || key == 'R' )
		reflMode = GL_REFLECTION_MAP_ARB;
	else
	if ( key == '0' )
		object = 0;
	else
	if ( key == '1' )
		object = 1;
	else
	if ( key == '2' )
		object = 2;
	else
	if ( key == '3' )
		object = 3;
	else
	if ( key == '4' )
		object = 4;
	else
	if ( key == '5' )
		object = 5;
	else
	if ( key == '6' )
		object = 6;
	else
	if ( key == '7' )
		object = 7;
	else
	if ( key == '8' )
		object = 8;
	else	
	if ( key == '+' )
	{
		if ( ++object > 8 )
			object = 0;
	}
	else
	if ( key == '-' )
	{
		if ( --object < 0 )
			object = 8;
	}
}

void	animate ()
{
	angle  = 0.04f * glutGet ( GLUT_ELAPSED_TIME );
	angle2 = 0.01f * glutGet ( GLUT_ELAPSED_TIME );

	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								//initialise glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );   
	glutInitWindowSize  ( 500, 500 );


								// create window
	int	glWin = glutCreateWindow ( "OpenGL cubemap example" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key );
	glutIdleFunc     ( animate );

	const char * vendor    = (const char *)glGetString ( GL_VENDOR     );
	const char * renderer  = (const char *)glGetString ( GL_RENDERER   );
	const char * version   = (const char *)glGetString ( GL_VERSION    );
	const char * extension = (const char *)glGetString ( GL_EXTENSIONS );

	printf ( "Vendor:   %s\nRenderer: %s\nVersion:  %s\n", vendor, renderer, version );

	if ( !isExtensionSupported ( "GL_ARB_texture_cube_map" ) )
	{
		printf ( "ARB_texture_cube_map NOT supported.\n" );

		return 1;
	}

	init ();

	const char * faces [6] = 
	{
		"cm_left.tga", 
		"cm_right.tga", 
		"cm_top.tga", 
		"cm_bottom.tga", 
		"cm_back.tga", 
		"cm_front.tga", 
	};

	texture = createCubeMap ( true, faces );

	glutMainLoop ();

	return 0;
}
